/***************************************
**  matrix.cpp                         **
****************************************
**  CPSC281             Assignment 2  **
**  February 13, 2000     Question 1  **
**  Ben Damm               230002428  **
**  Derek Lewis            230009120  **
***************************************/

#include <strstream>
#include <iomanip>

template<class T>
Matrix<T>::Matrix()
{
  rows = 1;
  cols = 1;
  data = new T[rows*cols];
}

template<class T>
Matrix<T>::Matrix(int _rows, int _cols)
{
  rows = _rows;
  cols = _cols;
  data = new T[rows*cols];
}

template<class T>
Matrix<T>::Matrix(const Matrix& matrix)
{
  rows = matrix.getRows();
  cols = matrix.getCols();
  data = new T[rows*cols];
  for (int row = 1; row <= getRows(); row++)
    for (int col = 1; col <= getCols(); col++)
      set(row,col, matrix.get(row,col) );
}

template<class T>
Matrix<T>::~Matrix()
{
  delete [] data;
}

template<class T>
void Matrix<T>::swap(Matrix& matrix)
{
  int _rows = matrix.getRows();
  int _cols = matrix.getCols();
  T* _data = new T[_rows*_cols];
  for (int row = 1; row <= _rows; row++)
    for (int col = 1; col <= _cols; col++)
      _data[(row-1)*_rows+(col-1)] = matrix.get(row,col);

  matrix.resize(getRows(), getCols());
  for (int row = 1; row <= getRows(); row++)
    for (int col = 1; col <= getCols(); col++)
      matrix.set(row,col, get(row,col) );

  resize(_rows,_cols);
  for (int row = 1; row <= _rows; row++)
    for (int col = 1; col <= _cols; col++)
      set(row,col, _data[(row-1)*_rows+(col-1)] );
}

template<class T>
Matrix<T>& Matrix<T>::operator=(const Matrix& matrix)
{
  Matrix tmp(matrix);
  swap(tmp);
  return *this;
}

template<class T>
void Matrix<T>::resize(int _rows, int _cols)
{
  delete [] data;
  rows = _rows;
  cols = _cols;
  data = new T[rows*cols];
}

template<class T>
int Matrix<T>::getRows() const
{
  return rows;
}

template<class T>
int Matrix<T>::getCols() const
{
  return cols;
}

template<class T>
void Matrix<T>::set(int row, int col, T val)
{
  if (row < 1 || row > getRows())
    throw MatrixException("set: OOB row");
  data[(row-1)*rows+(col-1)] = val;
}

template<class T>
T Matrix<T>::get(int row, int col) const
{
  if (row < 1 || row > getRows())
    throw MatrixException("get: OOB row");
  return data[(row-1)*rows+(col-1)];
}

template<class T>
Matrix<T> Matrix<T>::add(T val) const
{
  Matrix tmp(getRows(), getCols());
  for (int row = 1; row < getRows(); ++row)
    for (int col = 1; col < getCols(); ++col)
      tmp.set(row, col, get(row,col)+val);
  return tmp;    
}

template<class T>
void Matrix<T>::rowSwap(int row1, int row2)
{
  if (row1 < 1 || row1 > getRows() || row2 < 1 || row2 > getRows())
    throw MatrixException("rowSwap: OOB row");
  for (int col = 1; col <= getCols(); ++col)
  {
    T tmp = get(row1, col);
    set(row1, col, get(row2, col) );
    set(row2, col, tmp);
  }
}

template<class T>
void Matrix<T>::rowMultiply(int row, T val)
{
  if (row < 1 || row > getRows())
    throw MatrixException("rowMultiply: OOB row");
  for (int col = 1; col <= getCols(); ++col)
    set(row, col, get(row, col)*val );
}

template<class T>
void Matrix<T>::rowAdd(int row1, int row2, T val)
{
  if (row1 < 1 || row1 > getRows() || row2 < 1 || row2 > getRows())
    throw MatrixException("rowAdd: OOB row");
  for (int col = 1; col <= getCols(); ++col)
    set(row1, col, get(row1, col) + get(row2, col)*val );
}

template<class T>
bool Matrix<T>::equals(const Matrix& matrix) const
{
  if (getRows() != matrix.getRows() || getCols() != matrix.getCols())
    return false;
  for (int row = 1; row <= getRows(); row++)
    for (int col = 1; col <= getCols(); col++)
      if (get(row,col) != matrix.get(row,col))
        return false;
  return true;
}

template<class T>
bool Matrix<T>::isZero() const
{
  for (int row = 1; row <= getRows(); row++)
    for (int col = 1; col <= getCols(); col++)
      if (get(row,col) != 0)
        return false;
  return true;
}

template<class T>
bool Matrix<T>::isIdentity() const
{
  if (getRows() != getCols())
    return false;
  for (int row = 1; row <= getRows(); row++)
    for (int col = 1; col <= getCols(); col++)
      if (get(row,col) != 0 && !(row == col && get(row,col)==1))
        return false;
  return true;
}

template<class T>
Matrix<T> Matrix<T>::add(const Matrix& matrix) const
{
  if (getRows() != matrix.getRows() || getCols() != matrix.getCols())
    throw MatrixException("add: incompatible matrices");
  Matrix<T> ret(getRows(), getCols());
  for (int row = 1; row <= getRows(); ++row)
    for (int col = 1; col <= getCols(); ++col)
      ret.set(row, col, get(row, col)+matrix.get(row,col) );
  return ret;
}

template<class T>
Matrix<T> Matrix<T>::subtract(const Matrix& matrix) const
{
  if (getRows() != matrix.getRows() || getCols() != matrix.getCols())
    throw MatrixException("subtract: incompatible matrices");
  Matrix<T> ret(getRows(), getCols());
  for (int row = 1; row <= getRows(); ++row)
    for (int col = 1; col <= getCols(); ++col)
      ret.set(row, col, get(row, col)-matrix.get(row,col) );
  return ret;
}

template<class T>
Matrix<T> Matrix<T>::multiply(const Matrix& matrix) const
{
  if (getCols() != matrix.getRows())
    throw MatrixException("multiply: incompatible matrices");
  Matrix<T> ret(getRows(), getCols());
  for (int row = 1; row <= getRows(); ++row)
    for (int col = 1; col <= getCols(); ++col)
    {
      T tmp(0);
      for (int x = 1; x <= getCols(); ++x)
        tmp += get(row,x)*matrix.get(x,col);
      ret.set(row,col,tmp);
    }
  return ret;
}

template<class T>
Matrix<T> Matrix<T>::transpose() const
{
  Matrix tmp(getRows(), getCols());
  for (int row = 1; row <= getRows(); ++row)
    for (int col = 1; col <= getCols(); ++col)
      tmp.set(row,col, get(col,row) );
  return tmp;
}

template<class T>
ostream& Matrix<T>::print(ostream& out) const
{
  out << toString();
  
  return out;
}

template<class T>
string Matrix<T>::toString() const
{
  ostrstream out;

  out << "+";
  for (int col = 1; col <= getCols(); ++col)
    out << "---+";
   out << endl;
  for (int row = 1; row <= getRows(); ++row)
  {
    out << "|";
    for (int col = 1; col <= getCols(); ++col)
      out << setw(3) << get(row,col) << "|";
    out << endl << "+";
    for (int col = 1; col <= getCols(); ++col)
      out << "---+";
    if (row < rows)
      out << endl;
  }

  out << '\0';
  return out.str();
}

template<class T>
ostream& operator<<(ostream& out, Matrix<T>& matrix)
{
  matrix.print(out);
  return out;
}
